from copy import deepcopy
from os import environ
from pathlib import Path
from unittest.mock import ANY, Mock, patch

from django.template.library import InvalidTemplateLibrary

from coltrane import (
    DEFAULT_CACHES_SETTINGS,
    DEFAULT_COLTRANE_SETTINGS,
    DEFAULT_INSTALLED_APPS,
    DEFAULT_MIDDLEWARE,
    # _get_default_template_settings,
    initialize,
)
from coltrane.config.settings import get_config

base_dir = Path(".")
config = get_config()

DEFAULT_SETTINGS = {
    "BASE_DIR": base_dir,
    "ROOT_URLCONF": "coltrane.urls",
    "DEBUG": True,
    "SECRET_KEY": ANY,
    "INSTALLED_APPS": deepcopy(DEFAULT_INSTALLED_APPS),
    "CACHES": deepcopy(DEFAULT_CACHES_SETTINGS),
    "MIDDLEWARE": deepcopy(DEFAULT_MIDDLEWARE),
    "TEMPLATES": deepcopy(config.get_templates_settings()),
    "INTERNAL_IPS": [],
    "ALLOWED_HOSTS": [],
    "STATIC_ROOT": ANY,
    "STATIC_URL": ANY,
    "STATICFILES_DIRS": ANY,
    "TIME_ZONE": ANY,
    "LOGGING": ANY,
    "COLTRANE": deepcopy(DEFAULT_COLTRANE_SETTINGS),
    "SETTINGS_MODULE": "coltrane",
    "ENV": ANY,
}


def _get_settings():
    settings = deepcopy(DEFAULT_SETTINGS)

    settings["INSTALLED_APPS"].append("django_browser_reload")
    settings["MIDDLEWARE"].append("django_browser_reload.middleware.BrowserReloadMiddleware")

    return settings


def _get_settings_with_debug_false():
    settings = deepcopy(DEFAULT_SETTINGS)
    settings["DEBUG"] = False

    return settings


@patch("coltrane.is_whitenoise_installed", return_value=False)
@patch("coltrane.is_django_compressor_installed", return_value=False)
@patch("coltrane.is_django_unicorn_installed", return_value=False)
@patch("coltrane.is_unicorn_module_available", return_value=False)
@patch("coltrane._configure_settings")
def test_no_base_dir(_configure_settings, *args):
    initialize()

    expected = _get_settings()

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", return_value=False)
@patch("coltrane.is_django_compressor_installed", return_value=False)
@patch("coltrane.is_django_unicorn_installed", return_value=False)
@patch("coltrane.is_unicorn_module_available", return_value=False)
@patch("coltrane._configure_settings")
def test_with_base_dir(_configure_settings, *args):
    initialize(BASE_DIR=Path("test"))

    expected = _get_settings()
    expected["BASE_DIR"] = Path("test")
    expected["TEMPLATES"][0]["DIRS"] = [Path("test") / "templates"]

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", return_value=False)
@patch("coltrane.is_django_compressor_installed", return_value=False)
@patch("coltrane.is_django_unicorn_installed", return_value=False)
@patch("coltrane.is_unicorn_module_available", return_value=False)
@patch("coltrane._configure_settings")
def test_with_base_dir_as_string(_configure_settings, *args):
    initialize(BASE_DIR="test")

    expected = _get_settings()
    expected["BASE_DIR"] = Path("test")
    expected["TEMPLATES"][0]["DIRS"] = [Path("test") / "templates"]

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch(
    "coltrane.config.coltrane.configurators.templates.TemplatesConfigurator._get_template_tag_module_name",
    Mock(return_value="fake.templatetag"),
)
@patch("coltrane._configure_settings")
def test_with_template_tags(_configure_settings, tmp_path):
    (tmp_path / "templatetags").mkdir()
    # not actually used, but need the file here
    (tmp_path / "templatetags" / "sample_tag.py").touch()

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = ANY
    expected["TEMPLATES"][0]["DIRS"] = ANY
    expected["TEMPLATES"][0]["OPTIONS"]["builtins"].append("fake.templatetag")

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch(
    "coltrane.config.coltrane.configurators.templates.TemplatesConfigurator._get_template_tag_module_name",
    Mock(side_effect=InvalidTemplateLibrary),
)
@patch("coltrane._configure_settings")
def test_with_invalid_template_tag(_configure_settings, tmp_path):
    (tmp_path / "templatetags").mkdir()
    # not actually used, but need the file here
    (tmp_path / "templatetags" / "sample_tag.py").touch()

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = ANY
    expected["TEMPLATES"][0]["DIRS"] = ANY

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch(
    "coltrane.config.coltrane.configurators.templates.TemplatesConfigurator._get_template_tag_module_name",
    Mock(return_value="fake.templatetag"),
)
@patch("coltrane._configure_settings")
def test_with_template_tags_in_directory_with_py_extension(_configure_settings, tmp_path):
    (tmp_path / "templatetags").mkdir()
    (tmp_path / "templatetags" / "sample.py").mkdir()
    # not actually used, but need the file here
    (tmp_path / "templatetags" / "sample.py" / "sample_tag.py").touch()

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = ANY
    expected["TEMPLATES"][0]["DIRS"] = ANY
    expected["TEMPLATES"][0]["OPTIONS"]["builtins"].append("fake.templatetag")

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_debug_setting(_configure_settings):
    initialize(DEBUG=False)

    expected = _get_settings_with_debug_false()
    expected["TEMPLATES"] = ANY

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch.dict(environ, {"DEBUG": "False"})
@patch("coltrane._configure_settings")
def test_debug_env(_configure_settings):
    initialize()

    expected = _get_settings_with_debug_false()
    expected["TEMPLATES"] = ANY

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_time_zone_default(_configure_settings):
    initialize()

    expected = _get_settings()
    expected["TIME_ZONE"] = "UTC"

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_time_zone_setting(_configure_settings):
    initialize(TIME_ZONE="America/New_York")

    expected = _get_settings()
    expected["TIME_ZONE"] = "America/New_York"

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch.dict(environ, {"TIME_ZONE": "America/Chicago"})
@patch("coltrane._configure_settings")
def test_time_zone_env(_configure_settings):
    initialize()

    expected = _get_settings()
    expected["TIME_ZONE"] = "America/Chicago"

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_staticfiles_dirs(_configure_settings):
    initialize()

    expected = _get_settings()

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_staticfiles_dirs_content(_configure_settings, tmp_path: Path):
    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test.txt").write_text("test")

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = tmp_path
    expected["TEMPLATES"] = ANY
    expected["STATICFILES_DIRS"] = expected["STATICFILES_DIRS"] = [tmp_path]

    _configure_settings.assert_called_once_with(expected)


@patch("coltrane.is_whitenoise_installed", Mock(return_value=False))
@patch("coltrane.is_django_compressor_installed", Mock(return_value=False))
@patch("coltrane.is_django_unicorn_installed", Mock(return_value=False))
@patch("coltrane.is_unicorn_module_available", Mock(return_value=False))
@patch("coltrane._configure_settings")
def test_staticfiles_dirs_data(_configure_settings, tmp_path: Path):
    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text("{}")

    initialize(BASE_DIR=tmp_path)

    expected = _get_settings()
    expected["BASE_DIR"] = tmp_path
    expected["TEMPLATES"] = ANY
    expected["STATICFILES_DIRS"] = [tmp_path]

    _configure_settings.assert_called_once_with(expected)



from unittest.mock import MagicMock, patch

from coltrane.middleware import IsSecureMiddleware


def test_is_secure_false(rf):
    request = rf.get("/")
    assert request.is_secure() is False

    IsSecureMiddleware(MagicMock())(request)

    assert request.is_secure() is False


@patch("coltrane.config.settings.get_is_secure")
def test_is_secure_true(get_is_secure, rf):
    get_is_secure.return_value = True

    request = rf.get("/")
    assert request.is_secure() is False

    IsSecureMiddleware(MagicMock())(request)

    assert request.is_secure() is True



import logging
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import redirect_stdout
from io import StringIO
from multiprocessing import cpu_count
from pathlib import Path
from shutil import copy2
from types import SimpleNamespace

from django.conf import settings
from django.contrib.sitemaps.views import sitemap
from django.core import management
from django.core.management.base import BaseCommand
from halo import Halo  # type: ignore
from log_symbols.symbols import LogSymbols  # type: ignore

from coltrane.config.paths import (
    get_base_directory,
    get_extra_file_paths,
    get_output_directory,
    get_output_json,
    get_output_static_directory,
)
from coltrane.feeds import ContentFeed
from coltrane.manifest import Manifest, ManifestItem
from coltrane.module_finder import is_django_compressor_installed
from coltrane.renderer import StaticRequest
from coltrane.retriever import get_content_paths
from coltrane.urls import sitemaps
from coltrane.utils import threadpool

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Build all static HTML files and put them into a directory named output."

    is_force = False
    threads_count = 2
    manifest = None
    output_result_counts = SimpleNamespace(create_count=0, update_count=0, skip_count=0)
    request = StaticRequest("/")

    def add_arguments(self, parser):
        parser.add_argument(
            "--force",
            action="store_true",
            help="Force building all files",
        )

        parser.add_argument(
            "--threads",
            action="store",
            help="Number of threads to use when generating static files",
        )

        parser.add_argument(
            "--output",
            action="store",
            help="Output directory",
        )

        parser.add_argument(
            "--ignore",
            action="store_true",
            help="Ignore errors",
        )

    def _load_manifest(self) -> Manifest:
        return Manifest(manifest_file=get_output_json())

    def _generate_sitemap(self) -> None:
        if not self.output_directory:
            raise AssertionError("Missing output directory")

        sitemap_response = sitemap(request=self.request, sitemaps=sitemaps)
        sitemap_xml = sitemap_response.render().content.decode()
        (self.output_directory / "sitemap.xml").write_text(sitemap_xml)

    def _generate_rss(self) -> None:
        if not self.output_directory:
            raise AssertionError("Missing output directory")

        content_feed = ContentFeed()
        feed = content_feed.get_feed(None, request=self.request)  # type: ignore
        rss_xml = feed.writeString("utf-8")

        (self.output_directory / "rss.xml").write_text(rss_xml)

    @threadpool
    def _call_collectstatic(self) -> str:
        stdout = StringIO()
        stderr = StringIO()

        # Force DEBUG to always be `False` so that
        # whitenoise.storage.CompressedManifestStaticFilesStorage will use the static
        # assets with hashed filenames
        settings.DEBUG = False

        # TODO: Option to remove static files before re-generating
        management.call_command(
            "collectstatic",
            interactive=False,
            verbosity=1,
            stdout=stdout,
            stderr=stderr,
        )

        stderr.seek(0)

        # Get output from standard out and clean it up
        stdout.seek(0)
        collectstatic_stdout = stdout.read()
        collectstatic_stdout = collectstatic_stdout.replace(f" copied to '{get_output_static_directory()!s}'", "")[1:-2]

        collectstatic_stdout = f"Copy {collectstatic_stdout}"

        # TODO: Handle files in output.json that weren't
        # found in content? (--clean option?)

        return collectstatic_stdout

    @threadpool
    def _call_compress(self) -> str:
        stdout = StringIO()
        stderr = StringIO()

        # Force DEBUG to always be `False` so that
        settings.DEBUG = False
        settings.COMPRESS_OFFLINE = True

        # redirect_stdout is required
        # https://github.com/django-compressor/django-compressor/blob/develop/compressor/management/commands/compress.py#L385
        with redirect_stdout(stdout):
            management.call_command(
                "compress",
                verbosity=1,
                stdout=None,
                stderr=stderr,
            )

        compress_stdout = (
            stdout.getvalue().replace("Compressing... done\n", "").replace("Compressed ", "Compress ")[:-2]
        )

        return compress_stdout

    def _output_markdown_file(self, markdown_file: Path) -> None:
        if not self.manifest:
            raise AssertionError("Manifest must be loaded first")

        is_skipped = False

        item = ManifestItem.create(markdown_file)
        existing_item = self.manifest.get(markdown_file)

        if existing_item and not self.is_force:
            if item.mtime == existing_item.mtime:
                is_skipped = True
                self.output_result_counts.skip_count += 1
            elif item.md5 == existing_item.md5:
                # Update item in manifest to get newest mtime
                self.manifest.add(markdown_file)

                is_skipped = True
                self.output_result_counts.skip_count += 1

        if not is_skipped:
            if existing_item:
                self.output_result_counts.update_count += 1
            else:
                self.output_result_counts.create_count += 1

            rendered_html = item.render_html()

            item.generated_file_path.write_text(rendered_html)
            self.manifest.add(markdown_file)

    def _success(self, text: str, ending="\n") -> None:
        self.stdout.write(LogSymbols.SUCCESS.value, ending=" ")
        self.stdout.write(text, ending=ending)

    def _set_output_directory(self, options: dict) -> None:
        if options.get("output"):
            if not hasattr(settings, "COLTRANE"):
                settings.COLTRANE = {}

            if "OUTPUT" not in settings.COLTRANE:
                settings.COLTRANE["OUTPUT"] = {}

            settings.COLTRANE["OUTPUT"]["PATH"] = options["output"]

            # Override STATIC_ROOT if the output directory name is manually set
            try:
                settings.STATIC_ROOT = get_base_directory() / settings.COLTRANE["OUTPUT"]["PATH"] / "static"
            except KeyError:
                pass

            # Override STATIC_ROOT if the output directory is manually set
            try:
                settings.STATIC_ROOT = Path(settings.COLTRANE["OUTPUT"]["DIRECTORY"]) / "static"
            except KeyError:
                pass

    def handle(self, *args, **options):  # noqa: ARG002
        self.is_force: bool = False
        self.manifest: Manifest | None = None
        self.output_result_counts.create_count = 0
        self.output_result_counts.update_count = 0
        self.output_result_counts.skip_count = 0
        self.output_directory: Path | None = None
        self.errors: list[str] = []

        start_time = time.time()

        self.stdout.write(self.style.WARNING("Start generating the static site...\n"))

        spinner = Halo(spinner="dots")

        self._set_output_directory(options)

        collectstatic_future = self._call_collectstatic()

        compress_future = None

        if is_django_compressor_installed():
            compress_future = self._call_compress()

        self.output_directory = get_output_directory()
        self._success("Use output directory of ", ending="")
        self.stdout.write(self.style.WARNING(str(self.output_directory)))
        self.output_directory.mkdir(exist_ok=True)

        spinner.start("Load manifest")
        self.manifest = self._load_manifest()
        spinner.succeed()

        if "force" in options and options["force"] is True:
            self.is_force = True
            self._success("Force update because ", ending="")
            self.stdout.write(self.style.WARNING("--force"))

        spinner.start("Collect static files")
        collectstatic_stdout = collectstatic_future.result()
        spinner.succeed(collectstatic_stdout)

        if compress_future:
            spinner.start("Compress files")
            compress_stdout = compress_future.result()
            spinner.succeed(compress_stdout)

        spinner.start("Copy extra files")
        extra_file_count = 0
        for extra_file_path in get_extra_file_paths():
            copy2(extra_file_path, self.output_directory)
            extra_file_count += 1
        spinner.succeed(f"Copy {extra_file_count} extra files")

        if not self.is_force and self.manifest.static_files_manifest_changed:
            # At least one static file has changed, so re-render all files because
            # we don't have granularity to know which static files are used in
            # particular markdown or template files
            self.is_force = True
            self._success("Force update because static file(s) updated")

        if options.get("threads"):
            try:
                self.threads_count = int(options["threads"])
            except ValueError:
                pass
        else:
            try:
                self.threads_count = (cpu_count() // 2) - 1
            except Exception as ex:
                logger.exception(ex)

        spinner.start("Create HTML files")

        with ThreadPoolExecutor(max_workers=self.threads_count) as executor:
            logger.debug(f"Multithread with {self.threads_count} threads")
            pluralized_threads = "s" if self.threads_count > 1 else ""
            spinner.text = f"Create HTML files (use {self.threads_count} thread{pluralized_threads})"

            for path in get_content_paths(request=self.request):
                future = executor.submit(self._output_markdown_file, path)
                exception = future.exception()

                if exception:
                    error_detail = f"{exception.__class__.__name__}: {exception}"

                    if exception.__class__.__name__ == "FastDevVariableDoesNotExist":
                        error_detail = (
                            str(exception)
                            .replace("\n    ", ", ")
                            .replace(":\n, ", ": ")
                            .replace(
                                " does not exist in context.",
                                "' does not exist in template context.",
                            )
                        )[:-1]
                        error_detail = f"'{error_detail}"

                    error_message = f"Rendering {path} failed. {error_detail}"
                    self.errors.append(error_message)

        result_msg = f"Create {self.output_result_counts.create_count} HTML files, \
{self.output_result_counts.skip_count} unmodified, {self.output_result_counts.update_count} updated"

        spinner.succeed(result_msg)

        if self.manifest.is_dirty:
            spinner.start("Update sitemap.xml")
            self._generate_sitemap()
            spinner.succeed()

            spinner.start("Update rss.xml")
            self._generate_rss()
            spinner.succeed()

            spinner.start("Update output.json manifest")
            self.manifest.write_data()
            spinner.succeed()

        elapsed_time = time.time() - start_time

        for error_message in self.errors:
            spinner.fail(error_message)

        self.stdout.write("")

        if self.errors:
            self.stderr.write(self.style.ERROR(f"Static site output completed with errors in {elapsed_time:.4f}s\n"))

            if "ignore" not in options or not options["ignore"]:
                # Call sys.exit explicitly instead of CommandError for nicer error output
                sys.exit(1)
        else:
            self.stdout.write(self.style.SUCCESS(f"Static site output completed in {elapsed_time:.4f}s"))



from coltrane import _set_coltrane_setting


def test_set_coltrane_setting_missing_coltrane_in_settings():
    expected = {"COLTRANE": {"TITLE": "test title"}}

    settings = {}
    initialize_settings = {"COLTRANE_TITLE": "test title"}
    setting_name = "TITLE"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual


def test_set_coltrane_setting():
    expected = {"COLTRANE": {"TITLE": "test title"}}

    settings = {"COLTRANE": {}}
    initialize_settings = {"COLTRANE_TITLE": "test title"}
    setting_name = "TITLE"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual


def test_set_coltrane_setting_extra_file_names():
    expected = {"COLTRANE": {"EXTRA_FILE_NAMES": ["robots.txt"]}}

    settings = {"COLTRANE": {}}
    initialize_settings = {"COLTRANE_EXTRA_FILE_NAMES": "robots.txt"}
    setting_name = "EXTRA_FILE_NAMES"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual


def test_set_coltrane_setting_disable_wildcard_templates_true():
    expected = {"COLTRANE": {"DISABLE_WILDCARD_TEMPLATES": True}}

    settings = {"COLTRANE": {}}
    initialize_settings = {"COLTRANE_DISABLE_WILDCARD_TEMPLATES": True}
    setting_name = "DISABLE_WILDCARD_TEMPLATES"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual


def test_set_coltrane_setting_disable_wildcard_templates_true_string():
    expected = {"COLTRANE": {"DISABLE_WILDCARD_TEMPLATES": True}}

    settings = {"COLTRANE": {}}
    initialize_settings = {"COLTRANE_DISABLE_WILDCARD_TEMPLATES": "true"}
    setting_name = "DISABLE_WILDCARD_TEMPLATES"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual


def test_set_coltrane_setting_disable_wildcard_templates_not_true():
    expected = {"COLTRANE": {"DISABLE_WILDCARD_TEMPLATES": False}}

    settings = {"COLTRANE": {}}
    initialize_settings = {"COLTRANE_DISABLE_WILDCARD_TEMPLATES": 1}
    setting_name = "DISABLE_WILDCARD_TEMPLATES"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual


def test_set_coltrane_setting_is_secure_true():
    expected = {"COLTRANE": {"IS_SECURE": True}}

    settings = {"COLTRANE": {}}
    initialize_settings = {"COLTRANE_IS_SECURE": True}
    setting_name = "IS_SECURE"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual


def test_set_coltrane_setting_is_secure_true_string():
    expected = {"COLTRANE": {"IS_SECURE": True}}

    settings = {"COLTRANE": {}}
    initialize_settings = {"COLTRANE_IS_SECURE": "true"}
    setting_name = "IS_SECURE"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual


def test_set_coltrane_setting_is_secure_not_true():
    expected = {"COLTRANE": {"IS_SECURE": False}}

    settings = {"COLTRANE": {}}
    initialize_settings = {"COLTRANE_IS_SECURE": 1}
    setting_name = "IS_SECURE"
    actual = _set_coltrane_setting(settings, initialize_settings, setting_name)

    assert expected == actual



import logging
import sys
from contextlib import redirect_stdout
from copy import deepcopy
from io import StringIO
from os import environ, getcwd, getenv
from pathlib import Path
from typing import Any

from django import setup as django_setup
from django.conf import settings
from django.core.handlers.wsgi import WSGIHandler
from django.core.management import execute_from_command_line
from dotenv import load_dotenv

from coltrane.config.settings import DEFAULT_COLTRANE_SETTINGS, get_config
from coltrane.module_finder import (
    is_django_compressor_installed,
    is_django_unicorn_installed,
    is_unicorn_module_available,
    is_whitenoise_installed,
)
from coltrane.utils import dict_merge

logger = logging.getLogger(__name__)

__all__ = [
    "initialize",
    "run",
]


DEFAULT_CACHES_SETTINGS = {
    "default": {
        "BACKEND": "django.core.cache.backends.dummy.DummyCache",
    }
}

DEFAULT_MIDDLEWARE = [
    "coltrane.middleware.IsSecureMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.middleware.gzip.GZipMiddleware",
    "django.middleware.http.ConditionalGetMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]


DEFAULT_INSTALLED_APPS = [
    "django.contrib.humanize",
    "django.contrib.sitemaps",
    "django.contrib.staticfiles",
    "django_fastdev",
    "coltrane",
]

COLTRANE_SETTINGS_THAT_ARE_ARRAYS = ("EXTRA_FILE_NAMES",)
COLTRANE_SETTINGS_THAT_ARE_BOOLEANS = (
    "DISABLE_WILDCARD_TEMPLATES",
    "IS_SECURE",
    "DATA_JSON5",
)


def _get_current_command():
    if len(sys.argv) > 1:
        return sys.argv[1]


def _get_base_dir(base_dir: str | Path | None) -> Path:
    """
    Gets the base directory.
    """

    if base_dir is None:
        cwd = getcwd()

        base_dir = cwd / Path("sites")

        if not base_dir.exists():
            base_dir = cwd / Path("site")

            if not base_dir.exists():
                base_dir = Path(".")

    if isinstance(base_dir, str):
        base_dir = Path(base_dir)

    return base_dir


def _merge_installed_apps(django_settings: dict[str, Any], installed_apps: list[str]) -> list[str]:
    """
    Gets the installed apps from the passed-in settings and adds `coltrane` to it.
    """

    if "INSTALLED_APPS" in django_settings:
        installed_apps.extend(list(django_settings["INSTALLED_APPS"]))

        # Remove INSTALLED_APPS from django_settings since it overrides the default settings
        # that gets merged later
        del django_settings["INSTALLED_APPS"]

    return installed_apps


def _get_caches(django_settings: dict[str, Any]) -> dict:
    """Gets the configured cache. Defaults to the dummy cache."""

    caches = django_settings.get("CACHES", DEFAULT_CACHES_SETTINGS)

    if "default" not in caches:
        caches = DEFAULT_CACHES_SETTINGS

    if coltrane_cache := getenv("CACHE"):
        if coltrane_cache == "dummy":
            logging.info("Use dummy cache")

            caches["default"] = {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}
        elif coltrane_cache == "memory":
            logging.info("Use local memory cache")

            caches["default"] = {
                "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
                "LOCATION": getenv("CACHE_LOCATION", "unique-snowflake"),
            }
        elif coltrane_cache in ["filesystem", "redis", "memcache"]:
            if cache_locations := _get_from_env("CACHE_LOCATION"):
                cache_backend = None

                if coltrane_cache == "filesystem":
                    logging.info("Use filesystem cache")

                    cache_backend = "django.core.cache.backends.filebased.FileBasedCache"
                    cache_locations = cache_locations[0]
                elif coltrane_cache == "redis":
                    logging.info("Use redis cache")

                    cache_backend = "django.core.cache.backends.redis.RedisCache"
                elif coltrane_cache == "memcache":
                    logging.info("Use memcache cache")

                    cache_backend = "django.core.cache.backends.memcached.PyMemcacheCache"

                caches["default"] = {
                    "BACKEND": cache_backend,
                    "LOCATION": cache_locations,
                }
            else:
                raise Exception(f"Missing CACHE_LOCATION: '{coltrane_cache}'")
        else:
            raise Exception(f"Unknown cache backend: '{coltrane_cache}'")

    return caches


def _get_from_env(env_name: str) -> list[str]:
    """
    Retrieves environment value that could potentially be a list of strings.
    """

    env_values = []

    if value_from_env := getenv(env_name):
        env_values = value_from_env.split(",")

    return env_values


def _set_coltrane_setting(settings: dict, initialize_settings: dict, setting_name: str) -> dict:
    """
    Sets a setting on the `COLTRANE` dictionary that is in the environment or passed
    in to `initialize`. Environment takes precedence.

    For example:
    - `COLTRANE_TITLE=Awesome Blog 1` in `.env` equals `COLTRANE['TITLE'] = 'Awesome Blog 1'` in settings
    - `initialize(COLTRANE_TITLE='Awesome Blog 2')` equals `COLTRANE['TITLE'] = 'Awesome Blog 2'` in settings
    - `COLTRANE_TITLE=Awesome Blog 3` in `.env` and `initialize(COLTRANE_TITLE='Awesome Blog 4')` equals
        `COLTRANE['TITLE'] = 'Awesome Blog 3'` in settings
    """

    coltrane_setting_name = f"COLTRANE_{setting_name}"
    value = getenv(coltrane_setting_name) or initialize_settings.get(coltrane_setting_name)

    if value:
        # Make sure there is a `COLTRANE` setting
        if "COLTRANE" not in settings:
            settings["COLTRANE"] = {}

        if setting_name in COLTRANE_SETTINGS_THAT_ARE_ARRAYS:
            settings["COLTRANE"][setting_name] = value.split(",")
        elif setting_name in COLTRANE_SETTINGS_THAT_ARE_BOOLEANS:
            settings["COLTRANE"][setting_name] = False

            if isinstance(value, str):
                settings["COLTRANE"][setting_name] = value.lower() == "true"
            elif isinstance(value, bool):
                settings["COLTRANE"][setting_name] = value is True
        else:
            settings["COLTRANE"][setting_name] = value

    return settings


def _get_from_env_or_settings(django_settings: dict, key: str, default: Any) -> Any:
    """
    Get a setting from (in precedence order) the environment or the `COLTRANE` dictionary
    in settings. Normal `get_config_settings()` method does not work because it requires
    the Django settings to be configured.
    """

    val = default

    if setting_value := django_settings.get("COLTRANE", {}).get(key):
        val = setting_value

    if environment_value := getenv(f"COLTRANE_{key}"):
        val = environment_value

    return val


def _merge_settings(base_dir: Path, django_settings: dict[str, Any]) -> dict[str, Any]:
    """
    Merges the passed-in settings into the default `coltrane` settings.
    Passed-in settings will override the defaults.
    """

    # Assume that `argv[1] == "build"` means that the `build`
    # management command is currently being run
    is_build_management_command = _get_current_command() == "build"
    is_collectstatic_management_command = _get_current_command() == "collectstatic"

    debug = django_settings.get("DEBUG", getenv("DEBUG", "True") == "True")
    time_zone = django_settings.get("TIME_ZONE", getenv("TIME_ZONE", "UTC"))

    config = get_config(base_dir)

    staticfiles_dirs = []

    if not is_build_management_command and not is_collectstatic_management_command:
        # Add base directory for static files, the overriden static templatetag will handle the rest
        staticfiles_dirs = [base_dir]

    middleware = deepcopy(DEFAULT_MIDDLEWARE)
    installed_apps = deepcopy(DEFAULT_INSTALLED_APPS)

    if is_django_unicorn_installed():
        installed_apps.append("django_unicorn")

        if is_unicorn_module_available():
            installed_apps.append("unicorn")

    if is_whitenoise_installed():
        middleware.insert(1, "whitenoise.middleware.WhiteNoiseMiddleware")
        installed_apps.insert(0, "whitenoise.runserver_nostatic")

    if is_django_compressor_installed():
        installed_apps.append("compressor")

    if debug and not is_build_management_command:
        # Add settings required for django-browser-reload when appropriate
        middleware.append("django_browser_reload.middleware.BrowserReloadMiddleware")
        installed_apps.append("django_browser_reload")

        # Add content to "staticfiles" so django-browser-reload can monitor it
        content_directory = _get_from_env_or_settings(
            django_settings, "CONTENT_DIRECTORY", DEFAULT_COLTRANE_SETTINGS["CONTENT_DIRECTORY"]
        )

        # Add data to "staticfiles" so django-browser-reload can monitor it
        data_directory = _get_from_env_or_settings(
            django_settings, "DATA_DIRECTORY", DEFAULT_COLTRANE_SETTINGS["DATA_DIRECTORY"]
        )

        # Add content and data to staticfiles for each site so that they can be monitored by django-browser-reload
        for site in config.sites:
            site_content_directory = base_dir / "sites" / site.folder / content_directory

            if site_content_directory.exists():
                staticfiles_dirs.append(site_content_directory)

            site_data_directory = base_dir / "sites" / site.folder / data_directory

            if site_data_directory.exists():
                staticfiles_dirs.append(site_data_directory)

    templates = deepcopy(config.get_templates_settings())

    default_settings = {
        "BASE_DIR": base_dir,
        "ROOT_URLCONF": "coltrane.urls",
        "DEBUG": debug,
        "SECRET_KEY": getenv("SECRET_KEY"),
        "INSTALLED_APPS": _merge_installed_apps(django_settings, installed_apps),
        "CACHES": _get_caches(django_settings),
        "MIDDLEWARE": middleware,
        "TEMPLATES": templates,
        "INTERNAL_IPS": _get_from_env("INTERNAL_IPS"),
        "ALLOWED_HOSTS": _get_from_env("ALLOWED_HOSTS"),
        "STATIC_ROOT": base_dir / "output" / "static",
        "STATIC_URL": "static/",
        "STATICFILES_DIRS": staticfiles_dirs,
        "TIME_ZONE": time_zone,
        "LOGGING": {
            "version": 1,
            "disable_existing_loggers": False,
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                },
            },
            "root": {
                "handlers": ["console"],
                "level": "ERROR",
            },
        },
        "COLTRANE": DEFAULT_COLTRANE_SETTINGS,
        # Hard-coding `SETTINGS_MODULE` even though it kind of doesn't make sense, but
        # is needed for rendering a nice 500 page on local when debugging
        "SETTINGS_MODULE": "coltrane",
    }

    # Check for `COLTRANE` settings in env variables or passed into app.initialize()
    for setting_name in DEFAULT_COLTRANE_SETTINGS.keys():
        default_settings = _set_coltrane_setting(default_settings, django_settings, setting_name)

    if is_whitenoise_installed():
        default_settings["WHITENOISE_MANIFEST_STRICT"] = False
        default_settings["STATICFILES_STORAGE"] = "whitenoise.storage.CompressedManifestStaticFilesStorage"

    if is_django_compressor_installed():
        default_settings["COMPRESS_ENABLED"] = True

        default_settings["STATICFILES_FINDERS"] = (
            "django.contrib.staticfiles.finders.FileSystemFinder",
            "django.contrib.staticfiles.finders.AppDirectoriesFinder",
            "compressor.finders.CompressorFinder",
        )

        default_settings["TEMPLATES"][0]["OPTIONS"]["builtins"].append("compressor.templatetags.compress")

        if _get_current_command() == "compress":
            default_settings["COMPRESS_OFFLINE"] = True

    # Make sure BASE_DIR is a `Path` if it got passed in
    if "BASE_DIR" in django_settings and isinstance(django_settings["BASE_DIR"], str):
        django_settings["BASE_DIR"] = Path(django_settings["BASE_DIR"])

    # Override STATIC_ROOT if the output directory is manually set
    try:
        django_settings["STATIC_ROOT"] = Path(django_settings["COLTRANE"]["OUTPUT"]["DIRECTORY"]) / "static"
    except KeyError:
        pass

    django_settings = dict_merge(default_settings, django_settings, destination_overrides_source=True)
    logger.debug(f"Merged settings: {django_settings}")

    return django_settings


def _configure_settings(django_settings: dict[str, Any]) -> None:
    """
    Configures the settings in Django.
    """

    settings.configure(**django_settings)


def _load_environment_variables(django_settings: dict[str, Any]) -> None:
    load_dotenv(".env")

    if "ENV" not in django_settings:
        django_settings["ENV"] = {}

    django_settings["ENV"].update(dict(environ.items()))


def initialize(**django_settings) -> WSGIHandler:
    """
    Initializes the Django static site.
    """

    base_dir = _get_base_dir(django_settings.get("BASE_DIR"))
    _load_environment_variables(django_settings=django_settings)

    # coltrane_sites = django_settings.get("COLTRANE_SITES", {})

    django_settings = _merge_settings(base_dir, django_settings)
    _configure_settings(django_settings)

    django_setup()

    return WSGIHandler()


def run() -> None:
    """
    Run the Django management command based on `argv`.
    """

    if _get_current_command() == "compress":
        try:
            from compressor.exceptions import OfflineGenerationError
        except ImportError:
            logger.error("django-compressor is not installed.")
        else:
            try:
                stdout = StringIO()

                with redirect_stdout(stdout):
                    execute_from_command_line()

                compress_stdout = stdout.getvalue().replace("Compressing... done\n", "")
                print(compress_stdout)  # noqa: T201
            except OfflineGenerationError as e:
                if "No 'compress' template tags found in templates." in e.args[0]:
                    logger.error("No compress blocks found.")
                else:
                    raise
    else:
        execute_from_command_line()



#!/usr/bin/env python

from os import getcwd
from pathlib import Path

for i in range(1, 10000):
    file_name = str(i) + ".md"
    path = Path(getcwd()) / "content" / file_name
    # path.touch()
    path.write_text(
        """---
h1: Testing 1,2,3
numbers:
  - 1
  - 2
  - 3
number: 123456
test_string: this is a test string
---

# {{ h1 }}

This is a test

_ok_

**great**

## sub-heading

data.index.answer: {{ data.index.answer }}

data.index.author: {{ data.index.author }}

data.nested.more.another: {{ data.nested.more.another }}

Now: {{ now|date:"c" }}

For loop:

{% for i in numbers %}

- i: {{ i }}

{% endfor %}

Humanize built-in: {{ number|intcomma }}

Custom tag (should be "this is a string"): {{ test_string|cut_test:'test ' }}

Request: {{ request }}

Debug: {{ debug }}

![music]({% static 'images/music-note.svg' %})

"""
    )



from copy import deepcopy
from pathlib import Path

import pytest
from django.template import TemplateSyntaxError

from coltrane.config.settings import get_config


def test_include_md(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    # settings.TEMPLATES = deepcopy(_get_default_template_settings(tmp_path))
    coltrane = get_config(base_dir=tmp_path)
    settings.TEMPLATES = deepcopy(coltrane.get_templates_settings())

    (tmp_path / "templates").mkdir()
    (tmp_path / "templates/_partial.md").write_text("# partial")

    (tmp_path / "content").mkdir()
    (tmp_path / "content/index.md").write_text(
        """
# index

{% include_md '_partial.md' %}
"""
    )

    response = client.get("/")
    assert response.status_code == 200

    actual = response.content.decode()

    assert (
        """<h1 id="index">index</h1>
<p><h1 id="partial">partial</h1>"""
        in actual
    )


def test_include_md_loop(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    # settings.TEMPLATES = deepcopy(_get_default_template_settings(tmp_path))
    coltrane = get_config(base_dir=tmp_path)
    settings.TEMPLATES = deepcopy(coltrane.get_templates_settings())

    (tmp_path / "templates").mkdir()
    (tmp_path / "templates/_partial.md").write_text("# partial")

    (tmp_path / "content").mkdir()
    (tmp_path / "content/index.md").write_text(
        """
# index

{% for i in '01'|make_list %}
{% include_md '_partial.md' %}
{% endfor %}
"""
    )

    response = client.get("/")
    assert response.status_code == 200

    actual = response.content.decode()
    assert (
        """<h1 id="index">index</h1>
<p>
<h1 id="partial">partial</h1>


<h1 id="partial">partial</h1>

</p>"""
        in actual
    )


def test_include_md_metadata_in_template(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    # settings.TEMPLATES = deepcopy(_get_default_template_settings(tmp_path))
    coltrane = get_config(base_dir=tmp_path)
    settings.TEMPLATES = deepcopy(coltrane.get_templates_settings())

    (tmp_path / "templates").mkdir()
    (tmp_path / "templates/_partial.md").write_text(
        """
## partial

{{ variable }}

"""
    )

    (tmp_path / "content").mkdir()
    (tmp_path / "content/index.md").write_text(
        """---
variable: this-is-a-string-in-the-template
---

# index

{% include_md '_partial.md' %}
"""
    )

    response = client.get("/")
    assert response.status_code == 200

    actual = response.content.decode()
    assert (
        """<h1 id="index">index</h1>
<p><h2 id="partial">partial</h2>
<p>this-is-a-string-in-the-template</p>
</p>"""
        in actual
    )


def test_include_md_metadata_in_partial(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    coltrane = get_config(base_dir=tmp_path)
    settings.TEMPLATES = deepcopy(coltrane.get_templates_settings())

    (tmp_path / "templates").mkdir()
    (tmp_path / "templates/_partial.md").write_text(
        """---
request: this-is-a-string-in-the-partial
---

## partial

{{ request }}

"""
    )

    (tmp_path / "content").mkdir()
    (tmp_path / "content/index.md").write_text(
        """
# index

{% include_md '_partial.md' %}
"""
    )

    response = client.get("/")
    assert response.status_code == 200

    actual = response.content.decode()

    assert (
        """<h1 id="index">index</h1>
<p><h2 id="partial">partial</h2>
<p>this-is-a-string-in-the-partial</p>
</p>"""
        in actual
    )


def test_include_md_no_template(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    # settings.TEMPLATES = deepcopy(_get_default_template_settings(tmp_path))
    coltrane = get_config(base_dir=tmp_path)
    settings.TEMPLATES = deepcopy(coltrane.get_templates_settings())

    (tmp_path / "content").mkdir()
    (tmp_path / "content/index.md").write_text("{% include_md %}")

    with pytest.raises(TemplateSyntaxError):
        client.get("/")



from importlib.util import find_spec


def is_module_available(module_name: str) -> bool:
    """
    Helper function to check if a module is available.

    Could be an installed package or an available module.
    """

    return find_spec(module_name) is not None


def is_whitenoise_installed() -> bool:
    """
    Helper function to check if `whitenoise` is installed.
    """

    return is_module_available("whitenoise")


def is_django_compressor_installed() -> bool:
    """
    Helper function to check if `django-compressor` is installed.
    """

    return is_module_available("compressor")


def is_dj_angles_installed() -> bool:
    """
    Helper function to check if `dj_angles` is installed.
    """

    return is_module_available("dj_angles")


def is_django_unicorn_installed() -> bool:
    """
    Helper function to check if `django_unicorn` is installed.
    """

    return is_module_available("django_unicorn")


def is_unicorn_module_available() -> bool:
    """
    Helper function to check if there is a `unicorn` app available.
    """

    return is_module_available("unicorn")



from pathlib import Path
from unittest.mock import ANY

from django.utils.safestring import SafeString

from coltrane.renderer import StaticRequest
from coltrane.templatetags.coltrane_tags import directory_contents


def test_directory_contents(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")

    context = {"request": StaticRequest("/")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": ANY,
            "slug": "test",
            "now": ANY,
        }
    ]
    actual = directory_contents(context)

    assert actual == expected


def test_directory_contents_subdirectory_content(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test1").mkdir()
    (tmp_path / "content/test1/test2.md").write_text("test data")

    context = {"request": StaticRequest("/test1")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": ANY,
            "slug": "test1/test2",
            "now": ANY,
        }
    ]
    actual = directory_contents(context)

    assert actual == expected


def test_directory_contents_index_not_included(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test").mkdir()
    (tmp_path / "content/test/index.md").write_text("index")

    context = {"request": StaticRequest("/test")}
    expected = []
    actual = directory_contents(context)

    assert actual == expected


def test_directory_contents_explicit_directory(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test").mkdir()
    (tmp_path / "content/test/test.md").write_text("index")

    context = {"request": StaticRequest("/")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": ANY,
            "slug": "test/test",
            "now": ANY,
        }
    ]
    actual = directory_contents(context, directory="test")

    assert actual == expected


def test_directory_contents_title(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text(
        """---
title: this is a title
---

index
"""
    )

    context = {"request": StaticRequest("/")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": ANY,
            "slug": "test",
            "title": "this is a title",
            "now": ANY,
        }
    ]
    actual = directory_contents(context)

    assert actual == expected


def test_directory_contents_safe_string_subdirectory_content(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test1").mkdir()
    (tmp_path / "content/test1/test2.md").write_text("test data")

    context = {"request": StaticRequest("/test1")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": ANY,
            "slug": "test1/test2",
            "now": ANY,
        }
    ]
    actual = directory_contents(context, directory=SafeString("test1"))

    assert actual == expected


def test_directory_contents_exclude(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")
    (tmp_path / "content/another-test.md").write_text("another test")

    context = {"request": StaticRequest("/")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": ANY,
            "slug": "test",
            "now": ANY,
        }
    ]
    actual = directory_contents(context, exclude="another-test")

    assert actual == expected


def test_directory_contents_exclude_with_slash(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")
    (tmp_path / "content/another-test.md").write_text("another test")

    context = {"request": StaticRequest("/")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": ANY,
            "slug": "test",
            "now": ANY,
        }
    ]
    actual = directory_contents(context, exclude="/another-test")

    assert actual == expected


def test_directory_contents_exclude_with_comma_delimited_string(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")
    (tmp_path / "content/another-test.md").write_text("another test")
    (tmp_path / "content/yet-more-test.md").write_text("another test")

    context = {"request": StaticRequest("/")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": ANY,
            "slug": "test",
            "now": ANY,
        }
    ]
    actual = directory_contents(context, exclude="/another-test , yet-more-test")

    assert actual == expected


def test_directory_contents_order_by(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")
    (tmp_path / "content/another-test.md").write_text("another test")
    (tmp_path / "content/yet-more-test.md").write_text("another test")

    context = {"request": StaticRequest("/")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": None,
            "slug": "another-test",
            "now": ANY,
        },
        {
            "template": "coltrane/content.html",
            "toc": None,
            "slug": "test",
            "now": ANY,
        },
        {
            "template": "coltrane/content.html",
            "toc": None,
            "slug": "yet-more-test",
            "now": ANY,
        },
    ]
    actual = directory_contents(context, order_by="slug")

    assert actual == expected


def test_directory_contents_order_by_reversed(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")
    (tmp_path / "content/another-test.md").write_text("another test")
    (tmp_path / "content/yet-more-test.md").write_text("another test")

    context = {"request": StaticRequest("/")}
    expected = [
        {
            "template": "coltrane/content.html",
            "toc": None,
            "slug": "yet-more-test",
            "now": ANY,
        },
        {
            "template": "coltrane/content.html",
            "toc": None,
            "slug": "test",
            "now": ANY,
        },
        {
            "template": "coltrane/content.html",
            "toc": None,
            "slug": "another-test",
            "now": ANY,
        },
    ]
    actual = directory_contents(context, order_by="-slug")

    assert actual == expected



from unittest.mock import patch

from click.testing import CliRunner

from coltrane.console import cli


def test_record():
    runner = CliRunner()

    with patch("coltrane.console._run_management_command") as _run_management_command:
        result = runner.invoke(cli, ["record"])
        assert result.exit_code == 0

        _run_management_command.assert_called_once_with("build")


def test_build_alias():
    runner = CliRunner()

    with patch("coltrane.console._run_management_command") as _run_management_command:
        result = runner.invoke(cli, ["build"])
        assert result.exit_code == 0

        _run_management_command.assert_called_once_with("build")


def test_rec_alias():
    runner = CliRunner()

    with patch("coltrane.console._run_management_command") as _run_management_command:
        result = runner.invoke(cli, ["rec"])
        assert result.exit_code == 0

        _run_management_command.assert_called_once_with("build")


@patch("coltrane.console._run_management_command")
def test_record_force(_run_management_command):
    runner = CliRunner()
    runner.invoke(cli, ["record", "--force"])

    _run_management_command.assert_called_once_with("build", "--force")


@patch("coltrane.console._run_management_command")
def test_record_output(_run_management_command):
    runner = CliRunner()
    runner.invoke(cli, ["record", "--output", "public"])

    _run_management_command.assert_called_once_with("build", "--output", "public")


@patch("coltrane.console._run_management_command")
def test_record_threads(_run_management_command):
    runner = CliRunner()
    runner.invoke(cli, ["record", "--threads", "3"])

    _run_management_command.assert_called_once_with("build", "--threads", "3")


@patch("coltrane.console._run_management_command")
def test_record_ignore(_run_management_command):
    runner = CliRunner()
    runner.invoke(cli, ["record", "--ignore"])

    _run_management_command.assert_called_once_with("build", "--ignore")



import pytest

from coltrane.config.cache import ViewCache


def test_get_view_cache_seconds_none(settings):
    settings.COLTRANE = {"VIEW_CACHE": {"SECONDS": None}}
    view_cache = ViewCache()
    assert view_cache.seconds is None


def test_get_view_cache_seconds_configured(settings):
    settings.COLTRANE = {"VIEW_CACHE": {"SECONDS": 123}}
    view_cache = ViewCache()
    assert view_cache.seconds == 123


def test_get_view_cache_seconds_assert_not_enabled(settings):
    settings.COLTRANE = {}
    view_cache = ViewCache()

    with pytest.raises(AttributeError):
        view_cache.seconds  # noqa: B018



import logging

from django.contrib.sitemaps.views import _get_latest_lastmod, x_robots_tag
from django.contrib.sites.shortcuts import get_current_site
from django.core.paginator import EmptyPage, PageNotAnInteger
from django.http import FileResponse, Http404, HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import render
from django.template import TemplateDoesNotExist
from django.template.loader import select_template
from django.template.response import TemplateResponse
from django.utils.cache import patch_response_headers
from django.utils.http import http_date
from django.utils.timezone import now

from coltrane.config.cache import ViewCache
from coltrane.config.paths import get_file_path
from coltrane.config.settings import get_config, get_disable_wildcard_templates
from coltrane.renderer import MarkdownRenderer
from coltrane.retriever import get_data
from coltrane.sitemaps import ContentSitemap
from coltrane.wildcard_templates import get_potential_wildcard_templates

logger = logging.getLogger(__name__)


def _normalize_slug(slug: str) -> str:
    if slug is None:
        slug = ""

    if slug.endswith("/"):
        slug = slug[:-1]

    if slug.startswith("/"):
        slug = slug[1:]

    if slug == "":
        slug = "index"

    return slug


def _get_from_cache_if_enabled(slug: str) -> tuple[str, dict]:
    """
    Gets the slug from the cache if it's enabled.
    """

    template: str = ""
    context: dict = {}
    view_cache = ViewCache()

    if view_cache.is_enabled:
        cache_key = f"{view_cache.cache_key_namespace}{slug}"
        cached_value = view_cache.cache.get(cache_key)

        if cached_value:
            (template, context) = cached_value

    return (template, context)


def _set_in_cache_if_enabled(slug: str, template: str, context: dict) -> None:
    """
    Sets everything in the cache if it's enabled.
    """

    view_cache = ViewCache()

    if view_cache.is_enabled:
        cache_key = f"{view_cache.cache_key_namespace}{slug}"
        view_cache.cache.set(
            cache_key,
            (template, context),
            view_cache.seconds,
        )


def _render_markdown_for_potential_slugs(potential_slugs: list[str], request: HttpRequest):
    for slug in potential_slugs:
        try:
            (template, context) = MarkdownRenderer.instance().render_markdown(slug, request=request)

            return (template, context)
        except FileNotFoundError:
            pass

    raise FileNotFoundError()


def content(request: HttpRequest, slug: str = "index") -> HttpResponse:
    """Renders the markdown file stored in `content` or HTML template based on the slug from the URL.
    Adds data into the context from JSON files in the `data` directory.

    Will cache the rendered content if enabled.
    """

    logger.debug(f"request: {request}")

    config = get_config()
    logger.debug(f"config: {config}")

    site = config.get_site(request)
    logger.debug(f"site: {site}")

    template: str = ""
    context: dict = {}
    slug = _normalize_slug(slug)
    slug_with_index = f"{slug}/index"

    (template, context) = _get_from_cache_if_enabled(slug)
    set_in_cache = False

    try:
        if not template or not context:
            set_in_cache = True
            potential_slugs = []

            potential_slugs.extend([slug, slug_with_index])
            logger.debug(f"potential_slugs: {potential_slugs}")

            (template, context) = _render_markdown_for_potential_slugs(potential_slugs=potential_slugs, request=request)

            template = site.get_template_name(template_name=template, verify=True)
    except FileNotFoundError:
        potential_templates = []

        if site and site.is_custom:
            potential_templates.extend(
                [
                    f"{site.folder}/templates/{slug}.html",
                    f"{site.folder}/templates/{slug_with_index}.html",
                ]
            )

        # Typical templates based on the slug
        potential_templates.extend([f"{slug}.html", f"{slug_with_index}.html"])

        if get_disable_wildcard_templates() is False:
            potential_wildcard_templates = get_potential_wildcard_templates(slug)

            if site and site.is_custom:
                potential_wildcard_templates = [f"{site.folder}/templates/{t}" for t in potential_wildcard_templates]
                potential_templates.extend(potential_wildcard_templates)
            else:
                potential_templates.extend(potential_wildcard_templates)

        try:
            logger.debug(f"potential_templates: {potential_templates}")
            selected_template = select_template(potential_templates)

            logger.debug(f"selected_template: {selected_template}")
            template = selected_template.template.name
            logger.debug(f"template: {template}")
        except TemplateDoesNotExist:
            raise Http404(f"{slug} cannot be found") from None

        context.update(
            {
                "data": get_data(site=site),
                "slug": slug,
                "template": template,
                "now": now(),
            }
        )

    if set_in_cache:
        _set_in_cache_if_enabled(slug, template, context)

    context["site"] = str(site) if site else None

    logger.debug(f"template: {template}")

    response = render(
        request,
        template,
        context=context,
    )

    view_cache = ViewCache()

    if view_cache.is_enabled:
        patch_response_headers(response, cache_timeout=view_cache.seconds)

    return response


def file(request, file_name: str) -> FileResponse:
    """Serves a file from the file path.

    Based on code in https://adamj.eu/tech/2022/01/18/how-to-add-a-favicon-to-your-django-site/#what-the-file-type.
    """

    config = get_config()
    site = config.get_site(request)

    file_path = get_file_path(site=site, file_name=file_name)

    if not file_path.exists():
        raise Http404(f"{file_path.name} cannot be found")

    # Don't use context manager to open the file because it will be closed automatically
    # per https://docs.djangoproject.com/en/4.0/ref/request-response/#fileresponse-objects
    file = file_path.open("rb")

    return FileResponse(file)


def healthcheck(request) -> JsonResponse:  # noqa: ARG001
    """Serves a JSON response if the site is up and running."""

    return JsonResponse({"status": "ok"})


@x_robots_tag
def sitemap(request):
    """Returns a sitemap for all of the content for the current site.

    Based heavily on django.contrib.sitemaps.views.sitemap.
    """

    template_name = "sitemap.xml"
    content_type = "application/xml"

    req_protocol = request.scheme
    req_site = get_current_site(request)

    page = request.GET.get("p", 1)

    lastmod = None
    all_sites_lastmod = True
    urls = []

    try:
        config = get_config()
        site = config.get_site(request)

        content_sitemap = ContentSitemap()
        content_sitemap.site = site

        urls.extend(content_sitemap.get_urls(page=page, site=req_site, protocol=req_protocol))

        if all_sites_lastmod:
            site_lastmod = getattr(content_sitemap, "latest_lastmod", None)

            if site_lastmod is not None:
                lastmod = _get_latest_lastmod(lastmod, site_lastmod)
            else:
                all_sites_lastmod = False
    except EmptyPage as e:
        raise Http404(f"Page {page} empty") from e
    except PageNotAnInteger as e:
        raise Http404(f"No page '{page}'") from e

    # If lastmod is defined for all sites, set header so as
    # ConditionalGetMiddleware is able to send 304 NOT MODIFIED
    if all_sites_lastmod:
        headers = {"Last-Modified": http_date(lastmod.timestamp())} if lastmod else None
    else:
        headers = None

    return TemplateResponse(
        request,
        template_name,
        {"urlset": urls},
        content_type=content_type,
        headers=headers,
    )



from os import getenv
from pathlib import Path

import msgspec
from django.conf import settings

from coltrane.config.coltrane import Config
from coltrane.exceptions import ColtraneConfigParseError

# List of available `mistune` plugins: https://mistune.lepture.com/en/latest/plugins.html
DEFAULT_MISTUNE_PLUGINS = [
    "strikethrough",
    "footnotes",
    "table",
    "task_lists",
    "def_list",
    "abbr",
    "mark",
    "insert",
    "superscript",
    "subscript",
]

# Used to look at environment variables to merge into settings
DEFAULT_COLTRANE_SETTINGS = {
    "TITLE": "",
    "DESCRIPTION": "",
    "SITE_URL": "",
    "MARKDOWN_RENDERER": "mistune",
    "MISTUNE_PLUGINS": DEFAULT_MISTUNE_PLUGINS,
    "EXTRA_FILE_NAMES": [],
    "CONTENT_DIRECTORY": "content",
    "DATA_DIRECTORY": "data",
    "DISABLE_WILDCARD_TEMPLATES": False,
    "IS_SECURE": False,
    "DATA_JSON5": False,
    "SITES": {},
}


def get_coltrane_settings() -> dict:
    """
    Get the `COLTRANE` settings from the settings file.
    """

    if hasattr(settings, "COLTRANE"):
        if not isinstance(settings.COLTRANE, dict):
            raise TypeError("COLTRANE settings should be a dictionary")

        return settings.COLTRANE

    return DEFAULT_COLTRANE_SETTINGS


def get_markdown_renderer() -> str:
    """
    Get the markdown renderer. Defaults to "mistune".
    """

    markdown_renderer = get_coltrane_settings().get("MARKDOWN_RENDERER", "mistune")

    if markdown_renderer not in [
        "mistune",
    ]:
        raise AssertionError("Invalid markdown renderer")

    return markdown_renderer


def get_mistune_plugins() -> list[str]:
    """
    Get the `mistune` plugins.
    """

    return get_coltrane_settings().get("MISTUNE_PLUGINS", DEFAULT_MISTUNE_PLUGINS)


def get_site_url() -> str | None:
    """
    Get the configured site.
    """

    return get_coltrane_settings().get("SITE_URL")


def get_title() -> str | None:
    """
    Get the configured title.
    """

    return get_coltrane_settings().get("TITLE")


def get_description() -> str | None:
    """
    Get the configured description.
    """

    return get_coltrane_settings().get("DESCRIPTION")


def get_data_directory() -> str:
    """
    Get the configured data directory.
    """

    return get_coltrane_settings().get("DATA_DIRECTORY", DEFAULT_COLTRANE_SETTINGS["DATA_DIRECTORY"])


def get_content_directory() -> str:
    """
    Get the configured data directory.
    """

    return get_coltrane_settings().get("CONTENT_DIRECTORY", DEFAULT_COLTRANE_SETTINGS["CONTENT_DIRECTORY"])


def get_extra_file_names() -> list[str]:
    return get_coltrane_settings().get("EXTRA_FILE_NAMES", [])


def get_disable_wildcard_templates() -> bool:
    return get_coltrane_settings().get("DISABLE_WILDCARD_TEMPLATES", False)


def get_is_secure() -> bool:
    return get_coltrane_settings().get("IS_SECURE", False)


def get_data_json_5() -> bool:
    return get_coltrane_settings().get("DATA_JSON5", False)


# Global config object that is cached in the module
config: Config | None = None


def reset_config_cache():
    global config  # noqa: PLW0603

    config = None


def get_config(base_dir: Path | None = None) -> Config:
    global config  # noqa: PLW0603

    if config is not None:
        return config

    if base_dir is None:
        from django.conf import settings

        base_dir = settings.BASE_DIR

    if base_dir is None:
        raise AssertionError(f"Invalid base directory: {base_dir}")

    if isinstance(base_dir, str):
        base_dir = Path(base_dir)

    config_file_name = getenv("COLTRANE_CONFIG_FILE", "coltrane.toml")
    config_file_path = Path(config_file_name)

    potential_config_file_paths = [
        base_dir / "sites" / config_file_path,
        base_dir / "site" / config_file_path,
        base_dir / config_file_path,
    ]

    for path in potential_config_file_paths:
        if path.exists():
            try:
                config = msgspec.toml.decode(path.read_bytes(), type=Config)

                config.base_dir = base_dir
                config.config_file_name = config_file_name
                config.config_file_path = path
                break
            except msgspec.ValidationError as e:
                raise ColtraneConfigParseError(str(e)) from e

    if config is None:
        config = Config(base_dir=base_dir)

    return config



import pytest

from coltrane.renderer import MistuneMarkdownRenderer

parameters = [
    (
        """
# first header

## second header

### third header

# back to a header

more stuff here
""",
        """
<ul>
  <li>
    <a href="#first-header">first header</a>
    <ul>
      <li>
        <a href="#second-header">second header</a>
        <ul>
          <li><a href="#third-header">third header</a></li>
        </ul>
      </li>
    </ul>
  </li>
  <li>
    <a href="#back-to-a-header">back to a header</a>
  </li>
</ul>
""",
    ),
    # Not sure how this should function
    #     (
    #         """
    # ## second header
    # ## another second header
    # ### third header
    # # first header
    # more stuff here
    # """,
    #         """
    # <ul>
    #   <li>
    #     <a href="#second-header">second header</a>
    #   </li>
    #   <li>
    #     <a href="#another-second-header">another second header</a>
    #     <ul>
    #       <li><a href="#third-header">third header</a></li>
    #     </ul>
    #   </li>
    #   <li>
    #     <a href="#first-header">first header</a>
    #   </li>
    # </ul>
    # """,
    #     ),
    (
        """
# `code` header

more stuff here
""",
        """
<ul>
  <li>
    <a href="#code-header"><code>code</code> header</a>
  </li>
</ul>
""",
    ),
    (
        """
## second header

## another second header

### third header

#### fourth header

#### another fourth header

### more third header

more stuff here
""",
        """
<ul>
  <li>
    <a href="#second-header">second header</a>
  </li>
  <li>
    <a href="#another-second-header">another second header</a>
    <ul>
      <li>
        <a href="#third-header">third header</a>
        <ul>
          <li><a href="#fourth-header">fourth header</a></li>
          <li><a href="#another-fourth-header">another fourth header</a></li>
        </ul>
      </li>
      <li>
        <a href="#more-third-header">more third header</a>
      </li>
    </ul>
  </li>
</ul>
""",
    ),
    (
        """
# first

## second

### third

## more second

""",
        """
<ul>
  <li>
    <a href="#first">first</a>
    <ul>
      <li>
        <a href="#second">second</a>
        <ul>
          <li>
            <a href="#third">third</a>
          </li>
        </ul>
      </li>
      <li>
        <a href="#more-second">more second</a>
      </li>
    </ul>
  </li>
</ul>
""",
    ),
    (
        """
# first
""",
        """
<ul>
  <li>
    <a href="#first">first</a>
  </li>
</ul>
""",
    ),
    (
        """
## second
""",
        """
<ul>
  <li>
    <a href="#second">second</a>
  </li>
</ul>
""",
    ),
    (
        """
# first

## second
""",
        """
<ul>
  <li>
    <a href="#first">first</a>
    <ul>
      <li>
        <a href="#second">second</a>
      </li>
    </ul>
  </li>
</ul>
""",
    ),
    (
        """
# first

## second

# another first
""",
        """
<ul>
  <li>
    <a href="#first">first</a>
    <ul>
      <li>
        <a href="#second">second</a>
      </li>
    </ul>
  </li>
  <li>
    <a href="#another-first">another first</a>
  </li>
</ul>
""",
    ),
    (
        """
## first

### second

## another first
""",
        """
<ul>
  <li>
    <a href="#first">first</a>
    <ul>
      <li>
        <a href="#second">second</a>
      </li>
    </ul>
  </li>
  <li>
    <a href="#another-first">another first</a>
  </li>
</ul>
""",
    ),
    (
        """
# first

## second

### third

## another second
""",
        """
<ul>
  <li>
    <a href="#first">first</a>
    <ul>
      <li>
        <a href="#second">second</a>
        <ul>
          <li>
            <a href="#third">third</a>
          </li>
        </ul>
      </li>
      <li>
        <a href="#another-second">another second</a>
      </li>
    </ul>
  </li>
</ul>
""",
    ),
]


def eq(actual, expected):
    # actual = HTML(actual).prettify()
    # expected = HTML(expected).prettify()

    expected = expected.replace("\n", "").replace("  ", "")

    print("actual:")  # noqa: T201
    print(actual)  # noqa: T201
    print("expected:")  # noqa: T201
    print(expected)  # noqa: T201

    assert actual == expected


@pytest.mark.parametrize("text, expected", parameters)
def test_compare_toc(text, expected):
    markdown_renderer = MistuneMarkdownRenderer()

    (_, metadata) = markdown_renderer.render_markdown_text(text)
    actual = metadata["toc"]

    eq(actual, expected)



from os import environ

import pytest

from coltrane import DEFAULT_CACHES_SETTINGS, _get_caches
from tests.fixtures import *  # noqa: F403


@pytest.fixture
def caches_settings():
    return {"CACHES": {"default": {"test": "ok"}}}


def test_get_caches_default():
    expected = DEFAULT_CACHES_SETTINGS
    actual = _get_caches({})

    assert expected == actual


def test_get_caches_override_no_default():
    expected = DEFAULT_CACHES_SETTINGS
    actual = _get_caches({"CACHES": {"test": "no-default"}})

    assert expected == actual


def test_get_caches_override_default(caches_settings):
    expected = {"default": {"test": "ok"}}
    actual = _get_caches(caches_settings)

    assert expected == actual


def test_get_caches_env_dummy(env, caches_settings):
    environ["CACHE"] = "dummy"

    expected = {"default": {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}}
    actual = _get_caches(caches_settings)

    assert expected == actual


def test_get_caches_env_invalid(env, caches_settings):
    environ["CACHE"] = "invalid-cache-name"

    with pytest.raises(Exception) as e:
        _get_caches(caches_settings)

    assert e.exconly() == "Exception: Unknown cache backend: 'invalid-cache-name'"


class TestMemory:
    def test_get_caches(self, env, caches_settings):  # noqa: ARG002
        environ["CACHE"] = "memory"

        expected = {
            "default": {
                "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
                "LOCATION": "unique-snowflake",
            }
        }
        actual = _get_caches(caches_settings)

        assert expected == actual

    def test_get_caches_location(self, env, caches_settings):  # noqa: ARG002
        environ["CACHE"] = "memory"
        environ["CACHE_LOCATION"] = "frosty-snowflake"

        expected = {
            "default": {
                "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
                "LOCATION": "frosty-snowflake",
            }
        }
        actual = _get_caches(caches_settings)

        assert expected == actual


class TestFilesystem:
    def test_get_caches(self, env, caches_settings):  # noqa: ARG002
        environ["CACHE"] = "filesystem"
        environ["CACHE_LOCATION"] = "/root/tmp/var"

        expected = {
            "default": {
                "BACKEND": "django.core.cache.backends.filebased.FileBasedCache",
                "LOCATION": "/root/tmp/var",
            }
        }
        actual = _get_caches(caches_settings)

        assert expected == actual

    def test_get_caches_missing_location(self, env, caches_settings):  # noqa: ARG002
        environ["CACHE"] = "filesystem"

        with pytest.raises(Exception) as e:
            _get_caches(caches_settings)

        assert e.exconly() == "Exception: Missing CACHE_LOCATION: 'filesystem'"


class TestMemcache:
    def test_get_caches(self, env, caches_settings):  # noqa: ARG002
        environ["CACHE"] = "memcache"
        environ["CACHE_LOCATION"] = "127.0.0.1"

        expected = {
            "default": {
                "BACKEND": "django.core.cache.backends.memcached.PyMemcacheCache",
                "LOCATION": ["127.0.0.1"],
            }
        }
        actual = _get_caches(caches_settings)

        assert expected == actual

    def test_get_caches_missing_location(self, env, caches_settings):  # noqa: ARG002
        environ["CACHE"] = "memcache"

        with pytest.raises(Exception) as e:
            _get_caches(caches_settings)

        assert e.exconly() == "Exception: Missing CACHE_LOCATION: 'memcache'"


class TestRedis:
    def test_get_caches(self, env, caches_settings):  # noqa: ARG002
        environ["CACHE"] = "redis"
        environ["CACHE_LOCATION"] = "127.0.0.1"

        expected = {
            "default": {
                "BACKEND": "django.core.cache.backends.redis.RedisCache",
                "LOCATION": ["127.0.0.1"],
            }
        }
        actual = _get_caches(caches_settings)

        assert expected == actual

    def test_get_caches_missing_location(self, env, caches_settings):  # noqa: ARG002
        environ["CACHE"] = "redis"

        with pytest.raises(Exception) as e:
            _get_caches(caches_settings)

        assert e.exconly() == "Exception: Missing CACHE_LOCATION: 'redis'"



from datetime import datetime

from django.contrib.syndication.views import Feed

from coltrane.config.settings import get_config, get_description, get_site_url, get_title
from coltrane.retriever import ContentItem, get_content_items


class ContentFeed(Feed):
    title = get_title()
    description = get_description()
    _request = None

    def get_feed(self, obj, request):
        self._request = request

        return super().get_feed(obj, request)

    @property
    def site_url(self):
        site_url = get_site_url()

        if not site_url:
            raise AssertionError("COLTRANE_SITE_URL in .env or COLTRANE.SITE_URL in settings file is required")

        return site_url

    def items(self):
        site = None

        if self._request:
            site = get_config().get_site(self._request)

        return get_content_items(site=site)

    def item_title(self, item: ContentItem):
        return item.metadata.get("title")

    def item_description(self, item: ContentItem):
        return item.metadata.get("description")

    def item_link(self, item: ContentItem):
        site_url = self.site_url

        if site_url.endswith("/"):
            site_url = site_url[:-1]

        link = f"{site_url}{item.relative_url}"

        return link

    def item_pubdate(self, item: ContentItem) -> datetime | None:
        if publish_date := item.metadata.get("publish_date"):
            return publish_date

    def link(self, obj):  # noqa: ARG002
        return self.site_url



import importlib


def test_urlpatterns(settings):
    settings.DEBUG = False
    settings.COLTRANE["EXTRA_FILE_NAMES"] = []

    # Import module and reload it to get the right routes
    from coltrane import urls

    importlib.reload(urls)

    assert len(urls.urlpatterns) == 4

    assert urls.urlpatterns[0].pattern._route == "healthcheck"
    assert urls.urlpatterns[1].pattern._route == "sitemap.xml"
    assert urls.urlpatterns[2].pattern._route == "rss.xml"
    assert urls.urlpatterns[3].name == "content"


def test_urlpatterns_with_extra_file_names(settings):
    settings.DEBUG = False
    settings.COLTRANE["EXTRA_FILE_NAMES"] = ["robots.txt"]

    # Import module and reload it to get the right routes
    from coltrane import urls

    importlib.reload(urls)

    assert len(urls.urlpatterns) == 5

    assert urls.urlpatterns[0].pattern._route == "healthcheck"
    assert urls.urlpatterns[1].pattern._route == "sitemap.xml"
    assert urls.urlpatterns[2].pattern._route == "rss.xml"
    assert urls.urlpatterns[3].pattern._route == "robots.txt"
    assert urls.urlpatterns[4].name == "content"


def test_debug_urlpatterns(settings):
    settings.DEBUG = True
    settings.COLTRANE["EXTRA_FILE_NAMES"] = []

    # Import module and reload it to get the right routes
    from coltrane import urls

    importlib.reload(urls)

    assert len(urls.urlpatterns) == 5

    assert urls.urlpatterns[0].pattern._route == "__reload__/"
    assert urls.urlpatterns[1].pattern._route == "healthcheck"
    assert urls.urlpatterns[2].pattern._route == "sitemap.xml"
    assert urls.urlpatterns[3].pattern._route == "rss.xml"
    assert urls.urlpatterns[4].name == "content"



from pathlib import Path

from coltrane.config.paths import get_file_path


def test_get_file_path(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test.txt").write_text("test")

    expected = "test"

    actual = get_file_path("test.txt")
    assert actual.exists()

    assert expected == actual.read_text()


def test_get_file_path_not_exists(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()

    actual = get_file_path("test1234.txt")
    assert not actual.exists()



from pathlib import Path

from coltrane.config.coltrane import Site
from coltrane.retriever import get_data
from tests.fixtures import *  # noqa: F403


def test_get_data_directory(settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text('{"sample":1}')

    expected = {"test": {"sample": 1}}

    actual = get_data(site=default_site)

    assert actual == expected


def test_get_data_directory_invalid_json(settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text("")

    expected = {}
    actual = get_data(site=default_site)

    assert actual == expected


def test_get_data_directory_sub_directories(settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text('{"sample1":1}')
    (tmp_path / "data" / "another").mkdir()
    (tmp_path / "data" / "another" / "great.json").write_text('{"sample2":2}')
    (tmp_path / "data" / "another" / "more").mkdir()
    (tmp_path / "data" / "another" / "more" / "awesome.json").write_text('{"sample3":3}')

    expected = {
        "test": {"sample1": 1},
        "another": {"great": {"sample2": 2}, "more": {"awesome": {"sample3": 3}}},
    }
    actual = get_data(site=default_site)

    assert actual == expected


def test_get_data_directory_sub_directory_with_json(settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text('{"sample1":1}')

    expected = {
        "test": {"sample1": 1},
    }
    actual = get_data(site=default_site)

    assert actual == expected


def test_get_data_directory_sub_directory_with_json_utf8(settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text('{"sample1": "spræ"}')

    expected = {
        "test": {"sample1": "spræ"},
    }
    actual = get_data(site=default_site)

    assert actual == expected


def test_get_data_directory_sub_directory_with_json5(settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path
    settings.COLTRANE["DATA_JSON5"] = True

    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text('{"sample1":1,}')

    expected = {
        "test": {"sample1": 1},
    }
    actual = get_data(site=default_site)

    assert actual == expected


def test_get_data_directory_sub_directory_with_json5_utf8(settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path
    settings.COLTRANE["DATA_JSON5"] = True

    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test.json").write_text('{"sample1": "spræ",}')
    (tmp_path / "data" / "another.json").mkdir()

    expected = {
        "test": {"sample1": "spræ"},
    }
    actual = get_data(site=default_site)

    assert actual == expected


def test_get_data_directory_with_non_json_file(settings, tmp_path: Path, default_site: Site):
    settings.BASE_DIR = tmp_path

    (tmp_path / "data").mkdir()
    (tmp_path / "data" / "test1.json").write_text('{"sample":1}')
    (tmp_path / "data" / "test2.txt").write_text('{"sample":2}')

    expected = {"test1": {"sample": 1}}
    actual = get_data(site=default_site)

    assert actual == expected



from dataclasses import dataclass


@dataclass
class PathRanking:
    """Store a path with its score based on where the wildcard is in the path."""

    path: str

    def __init__(self, path: str):
        self.path = path
        self.score = self.score_path()

    def score_path(self):
        total_score = 0
        path = self.path.replace(".html", "")

        for idx, path_piece in enumerate(path.split("/")):
            path_idx = len(path.split("/")) - idx

            if path_piece == "*":
                # Inflate wildcards to ensure they are sorted as expected
                total_score += path_idx * 100
            else:
                total_score += path_idx

        return total_score

    def __str__(self):
        return f"{self.path} ({self.score})"


def _sort_potential_templates(template_paths):
    """Sort template paths based on where the wildcard is in the directory."""

    rankings = []

    for path in template_paths:
        rankings.append(PathRanking(path))

    rankings = sorted(rankings, key=lambda r: r.score, reverse=False)

    return [r.path for r in rankings]


def get_potential_wildcard_templates(slug: str) -> list[str]:
    """Get a list of potential wildcard HTML templates based on the slug."""

    wildcard_paths = []

    slug_pieces = slug.split("/")
    slug_pieces_count = len(slug_pieces)

    for outer_idx, _ in enumerate(slug_pieces):
        for inner_idx, _ in enumerate(slug_pieces):
            new_slug_pieces = []

            if outer_idx == inner_idx:
                new_slug_pieces.extend(slug_pieces[:outer_idx])
                new_slug_pieces.extend("*")
                new_slug_pieces.extend(slug_pieces[outer_idx + 1 :])

                wildcard_paths.append(new_slug_pieces)
            elif outer_idx < inner_idx and outer_idx > 0:
                new_slug_pieces.extend(slug_pieces[:outer_idx])
                new_slug_pieces.extend("*" * (slug_pieces_count - outer_idx))

                wildcard_paths.append(new_slug_pieces)
            elif outer_idx > inner_idx and (slug_pieces_count - outer_idx - 1) > 0:
                new_slug_pieces.extend("*" * (slug_pieces_count - outer_idx - 1))
                new_slug_pieces.extend(slug_pieces[outer_idx : outer_idx + 1])
                new_slug_pieces.extend("*" * (slug_pieces_count - outer_idx - 1))

                wildcard_paths.append(new_slug_pieces)

    # Add a catch-all for everything
    # Not needed if there are no potential sub-directories
    if slug_pieces_count > 1:
        new_slug_pieces = []

        for _ in range(slug_pieces_count):
            new_slug_pieces.append("*")

        wildcard_paths.insert(0, new_slug_pieces)

    potential_templates = []

    #  Convert the arrays of paths to paths and add to the list of potential templates
    for wildcard_option in wildcard_paths:
        wildcard_option_path = "/".join(wildcard_option)
        wildcard_option_path = f"{wildcard_option_path}.html"

        potential_templates.append(wildcard_option_path)

    potential_templates = _sort_potential_templates(potential_templates)

    return potential_templates



import importlib
from pathlib import Path

import pytest
from django.test import override_settings


@pytest.mark.urls("coltrane.urls")
def test_file_with_data(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.DEBUG = False
    settings.COLTRANE["EXTRA_FILE_NAMES"] = ["robots.txt"]

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "robots.txt").write_text("# allow all")

    # Import module and reload it to get the right routes
    from coltrane import urls

    importlib.reload(urls)

    with override_settings(ROOT_URLCONF=urls):
        response = client.get("/robots.txt")

    assert response.status_code == 200


def test_file_with_extra_file_names_setting(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.DEBUG = False
    settings.COLTRANE["EXTRA_FILE_NAMES"] = [
        "robots2.txt",
    ]

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "robots2.txt").write_text("# allow all")

    # Import module and reload it to get the right routes
    from coltrane import urls

    importlib.reload(urls)

    with override_settings(ROOT_URLCONF=urls):
        response = client.get("/robots2.txt")

    assert response.status_code == 200


def test_file_not_exists(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.DEBUG = True
    settings.COLTRANE["EXTRA_FILE_NAMES"] = ["robots.txt"]

    # Import module and reload it to get the right routes
    from coltrane import urls

    importlib.reload(urls)

    with override_settings(ROOT_URLCONF=urls):
        response = client.get("/robots.txt")

    assert response.status_code == 404

    # This will be robots.txt because the views.file route includes the
    # whole file name in the exception
    assert "robots.txt cannot be found" in response.content.decode()


@pytest.mark.urls("coltrane.urls")
def test_route_not_exists(client, settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.DEBUG = True
    settings.COLTRANE["EXTRA_FILE_NAMES"] = [
        "robots.txt",
    ]

    # Import module and reload it to get the right routes
    from coltrane import urls

    importlib.reload(urls)

    with override_settings(ROOT_URLCONF=urls):
        response = client.get("/robots1.txt")

    assert response.status_code == 404

    # This will be robots1 and not robots1.txt because the catch-all route for markdown
    # content only includes the "slug"
    assert "robots1 cannot be found" in response.content.decode()



from copy import deepcopy
from pathlib import Path

import pytest
from django.conf import settings

from coltrane.config.coltrane import Config
from coltrane.config.settings import DEFAULT_COLTRANE_SETTINGS, reset_config_cache


def get_coltrane_settings():
    coltrane_settings = deepcopy(DEFAULT_COLTRANE_SETTINGS)
    coltrane_settings["SITE_URL"] = "http://localhost"

    return coltrane_settings


def pytest_configure():
    base_dir = Path(".")
    config = Config(base_dir=base_dir)

    templates = deepcopy(config.get_templates_settings())

    settings.configure(
        BASE_DIR=base_dir,
        SECRET_KEY="this-is-a-secret",
        TEMPLATES=templates,
        ROOT_URLCONF="coltrane.urls",
        INSTALLED_APPS=[
            "django.contrib.sitemaps",
            "django_fastdev.apps.FastDevConfig",
            "coltrane",
        ],
        CACHES={
            "default": {
                "BACKEND": "django.core.cache.backends.dummy.DummyCache",
            }
        },
        STATIC_ROOT=base_dir / "output" / "static",
        SETTINGS_MODULE="coltrane",
        COLTRANE=get_coltrane_settings(),
    )


@pytest.fixture(autouse=True)
def reset_settings(settings):
    # Set the settings
    settings.COLTRANE = get_coltrane_settings()
    reset_config_cache()

    # Run test
    yield



from pathlib import Path

from coltrane.retriever import get_content_items


def test_location(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")

    actual = get_content_items()

    assert len(actual) == 1
    assert actual[0].relative_url == "/test"



from typing import TYPE_CHECKING, Optional

from django.contrib.sitemaps import Sitemap

from coltrane.retriever import ContentItem, get_content_items

if TYPE_CHECKING:
    from coltrane.config.coltrane import Site


class ContentSitemap(Sitemap):
    changefreq = "hourly"
    priority = 0.5
    site: Optional["Site"] = None

    def items(self):
        return get_content_items(site=self.site)

    def location(self, content_item: ContentItem) -> str:
        return content_item.relative_url



from dataclasses import dataclass

from django.core.cache import caches
from django.core.cache.backends.base import BaseCache

from coltrane.config.settings import get_coltrane_settings

AVAILABLE_CACHE_SETTINGS_KEYS = [
    "VIEW_CACHE",
    "DATA_CACHE",
]


@dataclass
class Cache:
    settings_key: str
    cache_key_namespace: str
    cache: BaseCache
    seconds: int
    is_enabled: bool = False

    def __init__(self, settings_key: str):
        self.settings_key = settings_key

        if self.settings_key not in AVAILABLE_CACHE_SETTINGS_KEYS:
            raise AssertionError("Invalid cache settings key")

        coltrane_settings = get_coltrane_settings()

        self.is_enabled = "SECONDS" in coltrane_settings.get(self.settings_key, {})

        if self.is_enabled:
            seconds = coltrane_settings[self.settings_key]["SECONDS"]

            if seconds is None:
                self.seconds = None
            else:
                self.seconds = int(seconds)

            self.cache_key_namespace = f"coltrane:{self.settings_key.lower()}:"

            cache_name = coltrane_settings[self.settings_key].get("CACHE_NAME", "default")
            self.cache = caches[cache_name]


@dataclass
class ViewCache(Cache):
    def __init__(self):
        super().__init__("VIEW_CACHE")


@dataclass
class DataCache(Cache):
    def __init__(self):
        super().__init__("DATA_CACHE")



import json
import logging
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

from django.http import HttpRequest

from coltrane.config.cache import DataCache
from coltrane.config.coltrane import Site
from coltrane.config.paths import get_content_directory, get_data_directory
from coltrane.config.settings import get_config, get_data_json_5
from coltrane.utils import dict_merge

logger = logging.getLogger(__name__)


MARKDOWN_EXTENSION_LENGTH = 3


def _add_data_from_path(data, data_directory, path):
    if path.is_file():
        directory_without_base_and_file_name = (str(path)).replace(str(data_directory), "").replace(path.name, "")

        # TODO: Check that .json5/.json is an extension first
        file_name = path.name.replace(".json5", "").replace(".json", "")
        value = None

        if get_data_json_5():
            try:
                import pyjson5

                try:
                    value = pyjson5.decode_buffer(path.read_bytes(), wordlength=0)
                except pyjson5.Json5DecoderException:
                    logger.exception(f"Invalid JSON5: '{file_name}'")
            except ImportError:
                pass
        else:
            try:
                value = json.loads(path.read_bytes())
            except json.decoder.JSONDecodeError:
                logger.exception(f"Invalid JSON: '{file_name}'")

        if value:
            new_data = {file_name: value}

            # For each part of the path between BASE_DIR/data and the JSON file,
            # add a new level (i.e. key) in the data dictionary; for example:
            # base_dir/data/some/new/test/here.json with {"one": "two"} ==
            # {"some": {"new": {"test": {"here": {"one": "two"}}}}}
            for key in reversed(directory_without_base_and_file_name.split("/")):
                if key:
                    new_data = {key: new_data}

            data = dict_merge(data, new_data)


def get_data(site: Site) -> dict:
    """
    Get and merge data from any JSON files recursively found in the `data` directory.
    """

    data = {}
    data_cache = DataCache()
    cache_key = ""

    if data_cache.is_enabled:
        cache_key = f"{data_cache.cache_key_namespace}data"
        data = data_cache.cache.get(cache_key, {})

        if data:
            return data

    data_directory = get_data_directory(site=site)

    for path in data_directory.rglob("*.json5"):
        _add_data_from_path(data, data_directory, path)

    for path in data_directory.rglob("*.json"):
        _add_data_from_path(data, data_directory, path)

    if data_cache.is_enabled:
        data_cache.cache.set(cache_key, data, timeout=data_cache.seconds)

    return data


def get_content_paths(
    request: HttpRequest | None = None, slug: str | None = None, site: Site | None = None
) -> Iterable[Path]:
    """
    Yield `Path`s for all markdown content in the content directory.
    """

    if not site and request:
        site = get_config().get_site(request)

    directory = get_content_directory(site=site)

    if slug:
        directory = directory / slug

    if not directory.exists():
        raise FileNotFoundError(f"Directory does not exist: {directory}")

    paths = directory.rglob("*.md")

    for path in paths:
        if path.is_file():
            yield path


@dataclass
class ContentItem:
    path: Path
    metadata: dict
    relative_url: str
    html: str


def get_content_items(site: Site | None = None, skip_draft: bool = True) -> Iterable[ContentItem]:  # noqa: FBT001, FBT002
    from coltrane.renderer import MarkdownRenderer

    paths = get_content_paths(site=site)
    _items = []

    content_directory = get_content_directory(site=site)
    content_directory_path_length = len(str(content_directory))

    for path in paths:
        (html, metadata) = MarkdownRenderer.instance().render_markdown_path(path)

        if skip_draft and metadata and "draft" in metadata and metadata["draft"] is True:
            continue

        path_str = str(path)
        relative_url = path_str[content_directory_path_length:-MARKDOWN_EXTENSION_LENGTH]

        if relative_url.endswith("/index"):
            relative_url = relative_url[:-6]

        content_item = ContentItem(path=path, metadata=metadata, html=html, relative_url=relative_url)
        _items.append(content_item)

    return _items



# Configuration file for the Sphinx documentation builder.

import toml

# -- Project information

project = "coltrane"
copyright = "2021, Adam Hill"  # noqa: A001
author = "Adam Hill"

pyproject = toml.load("../../pyproject.toml")
version = pyproject["project"]["version"]
release = version


# -- General configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "myst_parser",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx_inline_tabs",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

# -- Options for HTML output

html_theme = "furo"

# -- Options for EPUB output
epub_show_urls = "footnote"

autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 3

myst_heading_anchors = 3
myst_enable_extensions = ["linkify", "colon_fence"]



import json
from dataclasses import dataclass
from hashlib import md5 as md5_hash
from pathlib import Path

from django.template.loader import render_to_string

from coltrane.config.paths import get_output_directory, get_staticfiles_json
from coltrane.renderer import MarkdownRenderer, StaticRequest


@dataclass
class ManifestItem:
    """
    Stores information about a markdown file: the name, last modified time, and an MD5
    hash of the file contents.
    """

    _name: str
    _mtime: float
    _md5: str

    def __init__(self, name: str, mtime: float, md5: str):
        self._name = name
        self._mtime = mtime
        self._md5 = md5

    @property
    def slug(self) -> str:
        """
        The name minus the extension. Not exactly a slug because it could contain a
        forward slash for directories.
        """

        return self.name.replace(".md", "")

    @property
    def name(self) -> str:
        """
        The name of the markdown file. Does not contain the whole path, it would contain
        everything after the `content` directory. Does contain the extension.

        For example: the `name` of `/tmp/new-site/content/2022/new-file.md`
            would be `2022/new-file.md`.
        """

        return self._name

    @property
    def directory(self) -> str:
        """
        The directories before the file name.

        For example: the `path` of `/tmp/new-site/content/2022/new-file.md`
            would be `2022/`.
        """

        _directory = ""

        try:
            last_slash_idx = self.name.rindex("/")
            _directory = self.name[0:last_slash_idx]
        except ValueError:
            pass

        return f"/{_directory}"

    @property
    def mtime(self):
        """
        Last modified time of the file.
        """

        return self._mtime

    @property
    def md5(self):
        """
        MD5 hash of the file contents.
        """

        return self._md5

    @property
    def generated_file_path(self) -> Path:
        """
        The generated file path for the markdown file.
        """

        item_path = get_output_directory()

        for path in self.slug.split("/"):
            if path != "index":
                item_path = item_path / path
                item_path.mkdir(exist_ok=True)

        # index.md files be at the root as index.html
        if str(item_path).endswith("/index"):
            item_path = Path(str(item_path)[:-6])

        generated_file_path = item_path / "index.html"

        return generated_file_path

    @property
    def url_slug(self):
        url_slug = f"/{self.slug}"

        if url_slug.endswith("/index"):
            url_slug = url_slug[:-6]

        return url_slug

    def render_html(self):
        """
        Renders the markdown file into HTML.
        """

        # Mock an HttpRequest when generating the HTML for static sites
        request = StaticRequest(path=self.url_slug)

        (template, context) = MarkdownRenderer.instance().render_markdown(self.slug, request)
        rendered_html = render_to_string(template, context)

        return rendered_html

    @staticmethod
    def create(path: Path) -> "ManifestItem":
        """
        Initializes a new `ManifestItem` from a `Path`.
        """

        name = ManifestItem.get_name(path)
        mtime = path.stat().st_mtime
        md5 = md5_hash(path.read_bytes()).hexdigest()  # noqa: S324

        return ManifestItem(name=name, mtime=mtime, md5=md5)

    @staticmethod
    def get_name(path: Path) -> str:
        """
        Gets the `name` portion of the path, which is everything after the
        `content` directory.
        """

        name = ""

        if path.name == "staticfiles.json":
            return path.name

        # Assumes that there isn't a sub-folder called "content"
        for path_part in reversed(path.parts):
            if path_part == "content":
                break

            if path_part:
                if path_part == "/":
                    name = f"/{name}"
                elif name:
                    name = f"{path_part}/{name}"
                else:
                    name = path_part

        return name


@dataclass
class ManifestItems:
    """
    A store of all the markdown files in the manifest.
    """

    _data: dict[str, ManifestItem]

    def __init__(self):
        self._data = {}

    def get(self, name: str) -> ManifestItem:
        """
        Gets the markdown file information by name.
        """

        return self._data[name]

    def add(self, manifest_item: ManifestItem) -> None:
        self._data[manifest_item.name] = manifest_item

    def load(self, manifest_file: Path) -> None:
        """
        Retrieve the current manifest file (typically output.json) and store the data.
        """

        initial_data = json.loads(manifest_file.read_bytes())

        for key in initial_data.keys():
            values = initial_data[key]

            self._data[key] = ManifestItem(name=key, mtime=values.get("mtime"), md5=values.get("md5"))

    def __iter__(self):
        return iter(self._data.values())


@dataclass
class Manifest:
    """
    Represents a manifest file (output.json) which stores data for each markdown file
    that was output in the last build.
    """

    _manifest_file: Path
    _items: ManifestItems
    _static_files_manifest_changed: bool = False
    _is_dirty: bool = False

    def __init__(self, manifest_file: Path):
        self._manifest_file = manifest_file
        self._items = ManifestItems()

        if self._manifest_file.exists():
            self._items.load(manifest_file=manifest_file)

        staticfiles_manifest = get_staticfiles_json()

        if staticfiles_manifest.exists():
            if staticfiles_manifest_item := self.get(staticfiles_manifest):
                staticfiles_manifest_md5 = md5_hash(  # noqa: S324
                    staticfiles_manifest.read_bytes()
                ).hexdigest()

                if staticfiles_manifest_item.md5 != staticfiles_manifest_md5:
                    self.add(staticfiles_manifest)
                    self._static_files_manifest_changed = True
            else:
                self.add(staticfiles_manifest)
                self._static_files_manifest_changed = True

    @property
    def is_dirty(self):
        """
        Whether the manifest file has been changed.
        """

        return self._is_dirty

    @property
    def static_files_manifest_changed(self):
        """
        Whether the staticfiles manifest has been changed.
        """

        return self._static_files_manifest_changed

    def add(self, path: Path) -> ManifestItem:
        """
        Adds a path (normally a markdown file, but could also be `staticfiles.json`) to
        the manifest. Also used to update an existing file in the manifest.
        """

        item = ManifestItem.create(path)
        self._items._data[item.name] = item
        self._is_dirty = True

        return item

    def get(self, markdown_file: Path) -> ManifestItem | None:
        """
        Gets information about a markdown file from the manifest.
        """

        name = ManifestItem.get_name(markdown_file)

        try:
            return self._items.get(name)
        except KeyError:
            pass

        return None

    def write_data(self):
        """
        Writes the current manifest to the output file (typically output.json).
        """

        data = {}

        for item in self._items:
            data[item.name] = {"mtime": item.mtime, "md5": item.md5}

        self._manifest_file.write_text(json.dumps(data))



from pathlib import Path

from coltrane.sitemaps import ContentSitemap


def test_items(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")
    (tmp_path / "content/another-test.md").write_text("another test")

    actual = ContentSitemap().items()

    assert len(actual) == 2


def test_location(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/test.md").write_text("test data")

    items = ContentSitemap().items()
    actual = ContentSitemap().location(items[0])

    assert actual == "/test"


def test_location_index(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    (tmp_path / "content").mkdir()
    (tmp_path / "content/index.md").write_text("index")

    items = ContentSitemap().items()
    actual = ContentSitemap().location(items[0])

    assert actual == ""



from pathlib import Path

from coltrane.config.paths import get_extra_file_paths


def test_get_extra_file_paths(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.COLTRANE["EXTRA_FILE_NAMES"] = ["test.txt"]

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test.txt").write_text("test")

    actual = list(get_extra_file_paths())
    assert len(actual) == 1


def test_get_extra_file_paths_file_does_not_exist(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.COLTRANE["EXTRA_FILE_NAMES"] = ["test.txt"]

    actual = list(get_extra_file_paths())
    assert len(actual) == 0


def test_get_extra_file_paths_setting_is_empty(settings, tmp_path: Path):
    settings.BASE_DIR = tmp_path
    settings.COLTRANE["EXTRA_FILE_NAMES"] = []

    (tmp_path / "content").mkdir()
    (tmp_path / "content" / "test.txt").write_text("test")

    actual = list(get_extra_file_paths())
    assert len(actual) == 0



from copy import deepcopy
from pathlib import Path

from coltrane.config.settings import get_config


def _get_coltrane_toml() -> str:
    return """
[coltrane]

[[sites]]
folder = "example"
hosts = [
  "example.com"
]
"""


def _setup_sites_settings(settings, tmp_path: Path) -> None:
    settings.BASE_DIR = tmp_path / "sites"

    settings.BASE_DIR.mkdir()
    (settings.BASE_DIR / "coltrane.toml").write_text(_get_coltrane_toml())

    config = get_config()
    assert config.has_custom_sites

    # Set Django templates settings
    settings.TEMPLATES = deepcopy(config.get_templates_settings())


def _write_text(path: Path, text: str):
    (path.parent).mkdir(parents=True)
    (path).write_text(text)


def _client_get(client, url, headers: dict | None = None):
    if headers is None:
        headers = {"X-Forwarded-Host": "example.com"}

    return client.get(url, headers=headers)


def test_404(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    response = _client_get(client, "/")

    assert response.status_code == 404


def test_index_md(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    (settings.BASE_DIR / "example/content").mkdir(parents=True)
    (settings.BASE_DIR / "example/content/index.md").write_text("# index md")

    response = _client_get(client, "/")

    assert response.status_code == 200

    actual = response.content.decode()
    assert '<h1 id="index-md">index md</h1>' in actual


def test_md_without_slash(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    (settings.BASE_DIR / "example/content").mkdir(parents=True)
    (settings.BASE_DIR / "example/content/test-md-1.md").write_text("# test md 1")

    response = _client_get(client, "/test-md-1")

    assert response.status_code == 200

    actual = response.content.decode()
    assert '<h1 id="test-md-1">test md 1</h1>' in actual


def test_md_with_slash(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    (settings.BASE_DIR / "example/content").mkdir(parents=True)
    (settings.BASE_DIR / "example/content/test-md-2.md").write_text("# test md 2")

    response = _client_get(client, "/test-md-2/")

    assert response.status_code == 200

    actual = response.content.decode()
    assert '<h1 id="test-md-2">test md 2</h1>' in actual


def test_template_without_slash(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    (settings.BASE_DIR / "example/templates").mkdir(parents=True)
    (settings.BASE_DIR / "example/templates/test-template-1.html").write_text("Test template 1")

    response = _client_get(client, "/test-template-1")

    assert response.status_code == 200

    actual = response.content.decode()
    assert "Test template 1" == actual


def test_template_with_slash(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    (settings.BASE_DIR / "example/templates").mkdir(parents=True)
    (settings.BASE_DIR / "example/templates/test-template-2.html").write_text("Test template 2")

    response = _client_get(client, "/test-template-2/")

    assert response.status_code == 200

    actual = response.content.decode()
    assert "Test template 2" == actual


def test_md_with_template(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    (settings.BASE_DIR / "example/content").mkdir(parents=True)
    (settings.BASE_DIR / "example/content/test-md-with-template.md").write_text("""
---
template: template-1.html
---

# test md with template
""")

    (settings.BASE_DIR / "example/templates").mkdir(parents=True)
    (settings.BASE_DIR / "example/templates/template-1.html").write_text("""
template 1

{{ content }}
""")

    response = _client_get(client, "/test-md-with-template")

    assert response.status_code == 200

    actual = response.content.decode()
    assert "template 1" in actual
    assert '<h1 id="test-md-with-template">test md with template</h1>' in actual


def test_include(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    (settings.BASE_DIR / "example/templates").mkdir(parents=True)
    (settings.BASE_DIR / "example/templates/_partial.html").write_text("""
This is a partial
""")
    (settings.BASE_DIR / "example/templates/include-1.html").write_text("""
template 1

{% include '_partial.html' %}
""")

    response = _client_get(client, "/include-1")

    assert response.status_code == 200

    actual = response.content.decode()
    assert "This is a partial" in actual


def test_static(client, settings, tmp_path: Path):
    _setup_sites_settings(settings, tmp_path)

    text = """
static 1

<img src="{% static 'img/example.png' %}" />
"""
    _write_text(settings.BASE_DIR / "example/templates/static-1.html", text)

    actual = _client_get(client, "/static-1")

    assert 200 == actual.status_code
    assert '<img src="example/static/img/example.png" />' in actual.content.decode()



from importlib.metadata import version
from pathlib import Path
from unittest.mock import patch

from click.testing import CliRunner

from coltrane.console import APP_DIR, cli

FILES_PATH = Path(__file__).parent.parent.parent / "src/coltrane/default-files"


def test_create(tmp_path):
    runner = CliRunner()

    with patch("coltrane.console.get_random_secret_key") as get_random_secret_key:
        get_random_secret_key.return_value = "this-is-a-test"

        with runner.isolated_filesystem(temp_dir=tmp_path) as temp_dir:
            result = runner.invoke(cli, ["create"])
            assert result.exit_code == 0

            path_temp_dir = Path(temp_dir) / APP_DIR

            app_file = path_temp_dir / "app.py"
            assert app_file.exists()

            expected_app_py = (FILES_PATH / "app.py").read_text().replace("__coltrane_version__", version("coltrane"))
            assert app_file.read_text() == expected_app_py

            env_file = path_temp_dir / ".env"
            assert env_file.exists()
            assert env_file.read_text() == (FILES_PATH / "env").read_text() + "SECRET_KEY=this-is-a-test"

            content_dir = path_temp_dir / "content"
            assert content_dir.exists()

            assert (content_dir / "index.md").exists()
            assert (content_dir / "index.md").read_text() == "# index.md"

            data_dir = path_temp_dir / "data"
            assert data_dir.exists()

            assert (
                result.stdout
                == """Start creating files...

- Create .gitignore
- Create Dockerfile
- Create pyproject.toml
- Create README.md
- Create site/.env
- Create site/.watchmanconfig
- Create site/app.py
- Set site/app.py as executable
- Create site/gunicorn.conf.py
- Create site/content directory
- Create site/data directory
- Create site/static directory
- Create site/templates directory

For local development: coltrane play
Build static HTML: coltrane record
"""
            )


def test_create_existing_project(tmp_path):
    runner = CliRunner()

    with patch("coltrane.console.get_random_secret_key") as get_random_secret_key:
        get_random_secret_key.return_value = "this-is-a-test"

        with runner.isolated_filesystem(temp_dir=tmp_path):
            result = runner.invoke(cli, ["create"])
            assert result.exit_code == 0

            # Re-run create and make sure app.py check works
            result = runner.invoke(cli, ["create"])
            assert result.exit_code == 0

            assert (
                result.stdout
                == """Skip project creation because app.py already exists.

For local development: coltrane play
Build static HTML: coltrane record
"""
            )


def test_init_alias(tmp_path):
    runner = CliRunner()

    with patch("coltrane.console.get_random_secret_key") as get_random_secret_key:
        get_random_secret_key.return_value = "this-is-a-test"

        with runner.isolated_filesystem(temp_dir=tmp_path):
            result = runner.invoke(cli, ["init"])
            assert result.exit_code == 0

            assert "Start creating files..." in result.stdout



import msgspec

from coltrane.config.coltrane import Config


def test_empty():
    data = b""

    actual = msgspec.toml.decode(data, type=Config)

    assert actual
    assert len(actual.redirects) == 0
    assert len(actual.sites) == 1

    default_site = actual.sites[0]
    assert default_site.folder == ""
    assert len(default_site.hosts) == 1
    assert default_site.hosts[0] == "*"


def test_redirects():
    data = b"""
[[redirects]]
from_url = "/test-1"
to_url = "/"
permanent = false

[[redirects]]
from_url = "/test-2"
to_url = "/"
"""

    actual = msgspec.toml.decode(data, type=Config)

    assert actual
    assert actual.redirects
    assert len(actual.redirects) == 2


def test_sites():
    data = b"""
[[sites]]
folder = "adamghill"
hosts = [
  "0.0.0.0:80",
  "0.0.0.0:8020",
  "localhost:8020",
  "adamghill.localhost",
  "adamghill.com"
]

[[sites]]
folder = "alldjango"
hosts = [
  "0.0.0.0:8021",
  "localhost:8021",
  "alldjango.localhost",
  "alldjango.com"
]
"""

    actual = msgspec.toml.decode(data, type=Config)

    assert actual
    assert actual.sites
    assert len(actual.sites) == 2

    site_0 = actual.sites[0]
    assert site_0.folder == "adamghill"
    assert len(site_0.hosts) == 5

    site_1 = actual.sites[1]
    assert site_1.folder == "alldjango"
    assert len(site_1.hosts) == 4



from unittest.mock import patch

from click.testing import CliRunner

from coltrane.console import cli


def test_play():
    runner = CliRunner()

    with patch("coltrane.console._run_management_command") as _run_management_command:
        result = runner.invoke(cli, ["play"])
        assert result.exit_code == 0

        _run_management_command.assert_called_once_with("runserver", "0:8000")


def test_play_with_port():
    runner = CliRunner()

    with patch("coltrane.console._run_management_command") as _run_management_command:
        result = runner.invoke(cli, ["play", "--port=8001"])
        assert result.exit_code == 0

        _run_management_command.assert_called_once_with("runserver", "0:8001")


def test_serve_alias():
    runner = CliRunner()

    with patch("coltrane.console._run_management_command") as _run_management_command:
        result = runner.invoke(cli, ["serve"])
        assert result.exit_code == 0

        _run_management_command.assert_called_once_with("runserver", "0:8000")
